classdef PerfMeasure < handle
    % This class is a performance measure class for iterative methods.
    %
    % As functionality it has the following routines:
    %
    %   obj                         =   PerfMeasure(in,sSolvername);
    %   v                           =   PerfMeasure.matvec(u);  // v = A*u;
    %                                   PerfMeasure.stop();
    %                                   PerfMeasure.read(x,norm_r);
    %                                   PerfMeasure.resume();
    %   measure                     =   PerfMeasure.getMeasure();
    %   flag                        =   PerfMeasure.isTerminate();
    
    properties(Access=public)
        % solver name
        sSolvername
        % time measures
        delta_t         % total algorithm time
        delta_t_mv      % time exclusively for the matrix vector products
        t_last_toc      % stopwatch auxiliary parameter
        % system properties
        b               % right-hand side of the linear system
        A_fun           % callback for the matrix-vector product function
        % termination properties
        terminate
        % monitored properties
        array_matvec
        array_iterres
        array_trueres
        array_time
        array_time_mv
        arrayindex
        % % number of matrix vector products
        nummatvecs
        % % index of recycling point
        index_recycling
    end
    
    methods
        function [obj]      = PerfMeasure(in,sSolvername)
            % init
            obj.sSolvername     = sSolvername;
            % % init stopwatch (Matlab's builtin does not have a pause function)
            obj.delta_t         = 0;
            obj.delta_t_mv      = 0;
            obj.t_last_toc      = NaN;
            % % initialise system
            obj.b               = in.b;
            obj.A_fun           = in.A_fun;
            obj.terminate       = in.terminate;
            % % init convergence measure struct
            obj.array_matvec  	= NaN(1,2000);
            obj.array_iterres  	= NaN(1,2000);
            obj.array_trueres 	= NaN(1,2000);
            obj.array_time     	= NaN(1,2000);
            obj.array_time_mv  	= NaN(1,2000);
            obj.arrayindex      = 0;
            % % initilise numbers
            obj.nummatvecs      = 0;
            obj.index_recycling = [];
            % make first read for right-hand side b
            obj.read(0*in.b,norm(in.b));
            % % initialise clock
            tic;
        end
                
        function [v]        = matvec(obj,u)
            if(isnan(obj.t_last_toc))
                % timer was stopped
                % % time cost of the following computation is not monitored
                v               = obj.A_fun( u );
            else
                % timer was not stopped
                % % thus, the following computations shall be monitored
                obj.stop();
                t0              = toc;
                v               = obj.A_fun( u );
                t1              = toc;
                delta_t01       = t1 - t0;
                obj.nummatvecs  = obj.nummatvecs    + 1;
                % add delta_t01 to delta_t and delta_t_mv
                obj.delta_t     = obj.delta_t       + delta_t01;
                obj.delta_t_mv  = obj.delta_t_mv    + delta_t01;
                % continue with stopwatch
                obj.resume();
            end
        end
        
        function []         = stop(obj)
            if(isnan(obj.t_last_toc))
                % the timer is already stopped
            else
                % stop the timer
                % % read time
                this_time       = toc;
                % % add time minus start time to overall monitored time
                obj.delta_t    	= obj.delta_t + (this_time - obj.t_last_toc);
                % % set timer status to "stop"
                obj.t_last_toc  = NaN;
            end
        end
        
        function []         = read(obj,x,norm_r_iter)
            bool_is_timer_on    = ~isnan(obj.t_last_toc);
            obj.stop();
            % no matter if timer is stopped, the following computations
            % shall be done without measuring their computation time
            norm_r_true             = norm( obj.b - obj.A_fun(x) );
            obj.arrayindex          = obj.arrayindex + 1;
            i                       = obj.arrayindex;
            obj.array_matvec(i)     = obj.nummatvecs;
            obj.array_iterres(i)    = norm_r_iter;
            obj.array_trueres(i)    = norm_r_true;
            obj.array_time(i)       = obj.delta_t;
            obj.array_time_mv(i)    = obj.delta_t_mv;
            fprintf('Solver %s :\t%03.2f seconds total,\t%03.2f seconds in matvecs,\t#matvecs=%03i ,\t|res|=%01.3g .\n',...
                obj.sSolvername,obj.delta_t,obj.delta_t_mv,obj.nummatvecs,norm_r_iter);
            % the read function shall not change the status of the timer as
            % visible for the outside world
            if(bool_is_timer_on)
                obj.resume();
            end
        end
        
        function []         = resume(obj)
            % first stop to make sure we do not resume twice in a row
            obj.stop();
            % now do the computation for the actual resume
            obj.t_last_toc      = toc;
        end
        
        function [measure]  = getMeasure(obj)
            bool_is_timer_on    = ~isnan(obj.t_last_toc);
            obj.stop();
            % the following computations are done without stopping the time
            % that they require
            i                   = obj.arrayindex;   % lengths of the arrays
            measure.matvec      = obj.array_matvec( 1,1:i);
            measure.iterres    	= obj.array_iterres(1,1:i);
            measure.trueres   	= obj.array_trueres(1,1:i);
            measure.runtime   	= obj.array_time(   1,1:i);
            measure.runtime_mv	= obj.array_time_mv(1,1:i);
            measure.name        = obj.sSolvername;
            measure.recycling_index     = obj.index_recycling;
            % timer resumes if was activated before
            if(bool_is_timer_on)
                obj.resume();
            end
        end
        
        function []         = setRecycling(obj)
            bool_is_timer_on    = ~isnan(obj.t_last_toc);
            obj.stop();
            obj.index_recycling = obj.arrayindex;
            % timer resumes if was activated before
            if(bool_is_timer_on)
                obj.resume();
            end
        end
        
        function [flag]     = isTerminate(obj)
            bool_is_timer_on    = ~isnan(obj.t_last_toc);
            obj.stop();
            % the following computations are done without stopping the time
            % that they require
            i                   = obj.arrayindex;
            flag_tolabs         = (i>0) && (   obj.array_iterres(1,end) 	<= obj.terminate.tolabs);
            flag_maxmatvec      = obj.nummatvecs                            >= obj.terminate.maxmatvec;
            flag_maxruntime     = (i>0) && (obj.array_time(1,i)             >= obj.terminate.maxruntime);
            flag                = (flag_tolabs || flag_maxmatvec) || flag_maxruntime;
            % timer resumes if was activated before
            if(bool_is_timer_on)
                obj.resume();
            end
        end
        
    end
    
end