function [out] = Solver_GMsStab(in)
    % documentation of outputs:
    %   out.x                       :   numerical solution
    %   out.measure.matvec(i)    	:   number of matrix-vector products at i-th snap shot
    %   out.measure.iterres(i)   	:   norm of iteratively updated residual
    %   out.measure.trueres(i)     	:   norm of directly computed residual
    %   out.measure.runtime(i)     	:   time in seconds from start of method (excluded the time for computing the true residual)
    %   out.measure.runtime_mv(i)   :   time in seconds from start of method that is only required for matrix-vector products
    %   out.recyclingparam.hU     	:   matrix of auxiliary vectors
    %
    % inputs:
    %   in.A_fun                    :   linear operator from C^N --> C^N
    %   in.b                        :   right-hand side, s.t. A_fun(x) = b is to be solved for x
    %   in.x0                       :   initial guess for x
    %   in.terminate.tolabs         :   terminate when |b-A_fun(x)|<=tolabs
    %   in.terminate.maxmatvec      :   terminate when out.matvec(i)>=maxmatvec
    %   in.terminate.maxruntime     :   terminate when out.runtime(i)>=maxruntime
    %   in.param.s                  :   s<<N integer cut space dimension
    %   in.param.L                  :   L<<N integer polynomial dimension
    %   in.recyclingparam.tolabs2  	:   tolerance to write out hU
    %   in.recyclingparam.hU       	:   replaces search for first s start auxiliary vectors
    
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'hU') )
        fprintf('\n ---------- Using Solver GM(%i)stab with    recycling ---------- \n',in.param.s);
    else
        fprintf('\n ---------- Using Solver GM(%i)stab without recycling ---------- \n',in.param.s);
    end
    
    % init
    % % init performance measure
    myPerfMeasure    	= PerfMeasure(in,['GM(',num2str(in.param.s),')stab']);
    % % init system
    b                   = in.b;
    N                   = size(b,1);
    x                   = in.x0;
    A_fun               = @(v) myPerfMeasure.matvec(v);
    tolabs              = in.terminate.tolabs;
    % % init coefficients
    s                   = in.param.s;
    n2cycles            = 0;
    cRestart            = 1e-2;
    cReplace            = 1e-2;
    n2cyclesMax         = 3;
    % % init recycling
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'tolabs2'))
        tolabs2             = in.recyclingparam.tolabs2;
        hU                  = zeros(N,s,1);
    else
        tolabs2             = inf;
        hU                  = [];
    end
    % start method
    % % start performance monitoring
    myPerfMeasure.resume();
    % % init method
    % % % init flying restart
    xGlobal             = x;
    bLocal              = b - A_fun( x );
    r                   = bLocal;
    x                   = zeros(N,1);
    beta                = norm( r );
    betaLocal           = beta;
    % % % init projectors
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'hU') )
        rng('default');
        P                	= orth( randn(N,s) );
        V(:,:,1)        	= in.recyclingparam.hU;
        for q=1:s
            V(:,q,2)         	= A_fun( V(:,q,1) );
        end
        Z               	= tril( P' * V(:,:,2) );
        myPerfMeasure.stop();
        myPerfMeasure.read(xGlobal+x,beta);
        myPerfMeasure.resume();
        eta                 = P' * r(:,1);
       	[x,r]               = dirRbio(V,0,Z,x,r,eta);
        beta                = norm( r );
    else
        [P,V,Z,x,r,beta,myPerfMeasure] ...
                            = Initialisation(A_fun,x,r,beta,s,tolabs,myPerfMeasure);
    end
    myPerfMeasure.stop();
    myPerfMeasure.read(xGlobal+x,beta);
    myPerfMeasure.resume();
    betaMax             = max( beta, betaLocal );
    % run loop
    j=0;
    while(beta > tolabs)
        % - - - 1) Plan flying restart - - - 
        if(beta < cRestart * betaLocal)
            t_restart       = true;
            t_replace       = false;
        else
            t_restart       = false;
            if(beta < cReplace * betaMax)
                t_replace       = true;
            else
                t_replace       = false;
                betaMax         = max( betaMax, beta );
            end
        end
        % choose the cycle type
        if( t_restart || t_replace || n2cycles>n2cyclesMax )
            L               = 1;
        else
            L               = 2;
        end
        % - - - 2) Perform the cycle - - - 
%             % Preconditions
%             myPerfMeasure.stop();
%             err_Pr          = norm(   P' * r(:,1)               	,'inf')
%             err_V           = norm(   V(:,:,2) - A_fun( V(:,:,1) )  ,'inf')
%             err_Z           = norm(   Z - P' * V(:,:,2)   )
%             err_Lz          = norm(   Z - tril(Z)   )
%             err_res         = norm(   r(:,1) - ( b - A_fun( x ) )   )
%             myPerfMeasure.resume();
        if( L==1 )
            [Z,V,x,r,beta, myPerfMeasure]...
                            = GMstab1(A_fun,P,Z,V,x,r,beta,tolabs, myPerfMeasure,xGlobal);
            n2cycles        = 0;
            j               = j+1;
        else
            [Z,V,x,r,beta, myPerfMeasure]...
                            = GMstab2(A_fun,P,Z,V,x,r,beta,tolabs, myPerfMeasure,xGlobal);
            n2cycles        = n2cycles + 1;
            j               = j+2;
        end
        %if(beta >= tolabs2 && L==1)
        if(beta >= tolabs2)
            hU              = V(:,:,1);
            myPerfMeasure.setRecycling();
        end
        if( beta <= tolabs )
            break;
        end
        % - - - 3) Perform the flying restart - - - 
        if( t_restart || t_replace )
            disp('restart');
            r(:,1)          = bLocal - A_fun( x );
            eta             = P' * r(:,1);
            [x,r]          	= dirRbio(V,0,Z,x,r,eta);
            beta            = norm( r(:,1) );
            betaMax         = beta;
        else
            betaMax         = max( beta, betaMax );
        end
        if( t_restart )
            bLocal          = r(:,1);
            xGlobal         = xGlobal + x;
            x               = zeros(N,1);
            betaLocal       = beta;
        end
        myPerfMeasure.stop();
        myPerfMeasure.read(xGlobal + x,beta);
        if( myPerfMeasure.isTerminate() )
            break;
        end
        myPerfMeasure.resume();
    end
    myPerfMeasure.stop();
    % write output
    out.x               = x;
    out.measure         = myPerfMeasure.getMeasure();
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'tolabs2'))
        out.recyclingparam.hU	= hU;
    end
end