function [out] = Solver_IDRsStab_ref(in)
    % documentation of outputs:
    %   out.x                       :   numerical solution
    %   out.measure.matvec(i)    	:   number of matrix-vector products at i-th snap shot
    %   out.measure.iterres(i)   	:   norm of iteratively updated residual
    %   out.measure.trueres(i)     	:   norm of directly computed residual
    %   out.measure.runtime(i)     	:   time in seconds from start of method (excluded the time for computing the true residual)
    %   out.measure.runtime_mv(i)   :   time in seconds from start of method that is only required for matrix-vector products
    %
    % inputs:
    %   in.A_fun                    :   linear operator from C^N --> C^N
    %   in.b                        :   right-hand side, s.t. A_fun(x) = b is to be solved for x
    %   in.x0                       :   initial guess for x
    %   in.terminate.tolabs         :   terminate when |b-A_fun(x)|<=tolabs
    %   in.terminate.maxmatvec      :   terminate when out.matvec(i)>=maxmatvec
    %   in.terminate.maxruntime     :   terminate when out.runtime(i)>=maxruntime
    %   in.param.s                  :   s<<N integer cut space dimension
    %   in.param.L                  :   L<<N integer polynomial dimension
    
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'hU') )
        fprintf('\n ---------- Using Solver IDR(%i)stab(%i)ref with    recycling ---------- \n',in.param.s,in.param.L);
    else
        fprintf('\n ---------- Using Solver IDR(%i)stab(%i)ref without recycling ---------- \n',in.param.s,in.param.L);
    end
    
    % init
    % % init performance measure
    myPerfMeasure    	= PerfMeasure(in,['IDR(',num2str(in.param.s),')stab(',num2str(in.param.L),')ref']);
    % % init system
    b                   = in.b;
    N                   = size(b,1);
    x                   = in.x0;
    A_fun               = @(v) myPerfMeasure.matvec(v);
    tolabs              = in.terminate.tolabs;
    % % init storage
    s                   = in.param.s;
    L                   = in.param.L;
    % % init recycling
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'tolabs2') )
        tolabs2             = in.recyclingparam.tolabs2;
        hU                  = zeros(N,s,1);
    else
        tolabs2             = inf;
        hU                  = [];
    end
    % start method
    % % start performance monitoring
    myPerfMeasure.resume();
    % % init method
    r                   = b - A_fun( x );
    beta                = norm( r );
    % % % init projectors
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'hU') )
        rng('default');
        P                	= orth( randn(N,s) );
        V(:,:,1)        	= in.recyclingparam.hU;
        for q=1:s
            V(:,q,2)         	= A_fun( V(:,q,1) );
        end
        Z               	= tril( P' * V(:,:,2) );
        myPerfMeasure.stop();
        myPerfMeasure.read(x,beta);
        myPerfMeasure.resume();
        eta                 = P' * r(:,1);
       	[x,r]               = dirRbio(V,0,Z,x,r,eta);
        beta                = norm( r );
    else
        [P,V,Z,x,r,beta,myPerfMeasure] ...
                            = Initialisation(A_fun,x,r,beta,s,tolabs,myPerfMeasure);
    end
    myPerfMeasure.read(x,beta);
    % % init subsequent pages of the arrays
    V(:,:,2+[1:L])      = zeros(N,s,L);
    r(:,  1+[1:L])      = zeros(N,  L);
    myPerfMeasure.read(x,beta);
    % run loop
    while(true)
        % - - - Biorthogonalisation - - - 
        for k=0:1:(L-1)
            r(:,k+2)            = A_fun( r(:,k+1) );
            eta                 = P' * r(:,k+2);
            for q=1:s
                xi                  = Z \ eta;
                for g=-1:1:k
                    V(:,q,g+2)          = r(:,g+2) - [V(:,1:(q-1),g+3),V(:,q:s,g+2)] * xi;
                end
                V(:,q,g+3)          = A_fun( V(:,q,g+2) );
                Z(:,q)              = P' * V(:,q,g+3);
            end
            [x,r]               = dirRbio(V,k+1,Z,x,r,eta);
            myPerfMeasure.stop();
            norm_r          	= norm( r(:,1) );
            myPerfMeasure.read(x,norm_r);
            myPerfMeasure.resume();
        end
        % - - - Polynomial step - - - 
        beta                = norm( r(:,1) );
        [tau,beta]          = StabCoeffs(r,beta);
        x                   = x         + r(:,0+[1:L]) * tau;
        if( beta<=tolabs )
            myPerfMeasure.read(x,beta);
            break;
        end
        r(:,1)          	= r(:,1)    - r(:,1+[1:L]) * tau;
        V                   = permute(V,[1,3,2]);
        for q=1:s
            V(:,1,q)            = V(:,1,q)  - V(:,1+[1:L],q) * tau;
        end
        for q=1:s
            V(:,2,q)            = V(:,2,q)  - V(:,2+[1:L],q) * tau;
        end
        V                   = permute(V,[1,3,2]);
        %Z                   = -tau(L) * Z;
        Z                   = P' * V(:,:,2);
        myPerfMeasure.read(x,beta);
        if(beta >= tolabs2)
            hU              = V(:,:,1);
            myPerfMeasure.setRecycling();
        end
        if( myPerfMeasure.isTerminate() )
            break;
        end
    end
    myPerfMeasure.stop();
    % write output
    out.x               = x;
    out.measure         = myPerfMeasure.getMeasure();
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'tolabs2') )
        out.recyclingparam.hU    = hU;
    end
end